/*

ZZipLite is my redesign, recoding and adaptation for usage in Delphi of
ZZip library for lossless, block-sorting data compression by Damien Debin

Recoded by A.Myasnikov

e-mail: alexanderwdark@ya.ru

*/

#include <stdlib.h>
#include <memory.h>


#define MAX_FILES			65535
#define RUN_LENGTH_MAX		512
#define FILENAME_LENGTH_MAX (256+4+1)
#define DEFAULT_BLOCK_SIZE	(1024*1024)

#define WIN32_ASM_CALL		0xE8
#define ASCII_H_E			0x6568
#define ASCII_CR			10
#define ASCII_SP			32
#define TAG_CAPS			30

#define ROUND32(a)		(((uint32)(a+32))&(~31UL))

#define MEDIAN(a,b,c)	((a)<(b))?(((b)<(c))?(b):(((a)<(c))?(c):(a))):(((a)<(c))?(a):(((b)<(c))?(c):(b)))


# define MAX(a,b)		(((a)>(b))?(a):(b))
# define MIN(a,b)		(((a)<(b))?(a):(b))
# define GET_TSC(b)
# define STAT_ADD_TIME(a,b,c)
# define STAT_ADD_SIZE(a,b)


typedef enum
{
  false = 0,
  true  = 1
} bool;


typedef unsigned char  uint8;
typedef unsigned short uint16;
typedef unsigned int   uint32;
typedef   signed char  sint8;
typedef   signed short sint16;
typedef   signed int   sint32;
typedef   signed int   sint;
typedef   signed long  slong;
typedef unsigned int   uint;


typedef unsigned __int64 uint64;
typedef   signed __int64 sint64;



int last_error;

typedef union
  {
    uint64 u64;
    struct
      {
        uint32 l;
        uint32 h;
      } d;
  } uuint64;

typedef union
  {
    sint64 s64;
    struct
      {
        uint32 l;
        uint32 h;
      } d;
  } ssint64;

typedef union
  {
    uint32 u32;
    struct
      {
        uint16 l;
        uint16 h;
      } w;
    struct
      {
        uint8 ll;
        uint8 lh;
        uint8 hl;
        uint8 hh;
      } b;
  } uuint32;

typedef enum
{
  NO_TYPE,
  TEXT,
  MULTIMEDIA,
  BIN,
  WIN_EXE
} block_types;

typedef enum
{
  NONE,
  CREATE,
  EXTRACT,
  TEST,
  LIST,
  UPDATE,
  DELETE
} actions;

typedef enum
{
  UNCOMPRESS,
  TEST_CRC
} modes;

typedef enum
{
  OK                       =  0,
  NOT_ENOUGH_MEMORY        = -1,
  CRC_ERROR                = -2,
  TOO_MANY_SESSION         = -3,
  BAD_PARAMETER            = -4,
  UNEXPECTED_EOF           = -5,
  NOT_A_ZZIP_FILE          = -6,
  UNSUPPORTED_VERSION      = -7,
  CANNOT_CLOSE_INPUT_FILE  = -8,
  CANNOT_CLOSE_OUTPUT_FILE = -9,
  CANNOT_OPEN_INPUT_FILE   = -10,
  CANNOT_OPEN_OUTPUT_FILE  = -11,
  FSEEK_INPUT_FILE         = -12,
  FSEEK_OUTPUT_FILE        = -13,
  FILE_NOT_FOUND           = -14
} errors;

/*
 * Archive header
 * 2 bytes : "ZZ"
 * 1 byte  : Zzip version number used for compression
 * 4 bytes : number of files in the archive
 */

typedef struct
  {
    char   magic[2];
    uint8  version_number;
    uint32 nb_of_file;
  } header_arc_s;

/*
 * File header
 * 2 bytes : attributes
 * 2 bytes : number of blocks
 * 4 bytes : pointer to name (without trailing '\0')
 * 4 bytes : date/time of modification
 * 4 bytes : file size = size of all blocks
 * 4 bytes : original file size
 */




typedef struct
  {
    uint8		mtf_max_char;
    bool		multimedia_test;
    bool		rle_encoding;
    bool		english_encoding;
    uint		compression_mode;
    uint		mm_type;
    sint		nblock;
    uint32		crc;
    block_types	type;
    uint8		*buffer;
    uint8		*buffer_length;
  } block_param_s;

block_param_s	block;


typedef struct
  {
    bool			multimedia_test;
    bool			with_path;
    uint			compression_mode;
    uint32			block_size;
    uint32			input_total;
    uint32			output_total;


    block_types		type;
    actions			action;

  } session_param_s;


typedef struct
  {
    bool   t_stamp;
    uint32 cpuspeed;
    uint32 time_ana;
    uint32 time_rle;
    uint32 time_txt;
    uint32 time_mm;
    uint32 time_bwt1;
    uint32 time_bwt2;
    uint32 time_bwt3;
    uint32 time_bwt4;
    uint32 time_mtf;
    uint32 time_st0;
    uint32 time_st1;
    uint32 time_tot;
    uint32 kb_ana;
    uint32 kb_rle;
    uint32 kb_txt;
    uint32 kb_mm;
    uint32 kb_bwt;
    uint32 kb_mtf;
    uint32 kb_st0;
    uint32 kb_st1;
    uint32 kb_tot;
  } time_stat_s;


# define ALIGN

# define START_OFFSET 0L

#define WRITE_M(a,b)	{ memcpy(buffer_in, (a), (b)); buffer_in += (b); }
#define READ_M(a,b)		{ memcpy((a), buffer_in, (b)); buffer_in += (b); }
#define IO_ERROR()		{ last_error = errno; return; }
#define CHECK_IO_R()	{ if (ferror(session->input_file) != 0) last_error = errno; if (feof(session->input_file) != 0) last_error = UNEXPECTED_EOF; }
#define CHECK_IO_W()	{ if (ferror(session->output_file) != 0) last_error = errno; }
#define FTELL_I(a)		{ if ((a = ftell(session->input_file )) == -1) last_error = errno; }
#define FTELL_O(a)		{ if ((a = ftell(session->output_file)) == -1) last_error = errno; }
#define FSEEK_I(a,b)	{ if (fseek(session->input_file, (a), (b)) != 0)  last_error = FSEEK_INPUT_FILE; }
#define FSEEK_O(a,b)	{ if (fseek(session->output_file, (a), (b)) != 0) last_error = FSEEK_OUTPUT_FILE; }

int last_error = OK;

block_param_s block = { 0, false, false, false, 0, 0, 0, 0, NO_TYPE, NULL, NULL };

static session_param_s *session = NULL;
static union
  {
    uint8  *buffer8;
    uint16 *buffer16;
    uint32 *buffer32;
  } mem = { NULL };


# define NB_MAX_SESSION 8


# define INLINE   __inline

INLINE static
void *MyFree(void *pp)
{
  free(pp);
  return NULL;
}

INLINE static
void *MyMalloc(size_t size)
{
  void *mm = malloc(size);
  if (mm == NULL) last_error = NOT_ENOUGH_MEMORY;
  return mm;
}

INLINE static
void *MyRealloc(void   *pp,
                size_t size)
{
  void *mm = realloc(pp, size);
  if (mm == NULL) last_error = NOT_ENOUGH_MEMORY;
  return mm;
}

void CleanMemory()
{
  mem.buffer8 = MyFree(mem.buffer8);
}



#define STAT_SIZE			(8*1024)
#define NB_S				64
#define BLOCKSTAT_THRESHOLD	5400



#define MULT 2


#define CODE_VALUE_BITS     16
#define TOP_VALUE           ((1UL << CODE_VALUE_BITS) - 1)
#define FIRST_QTR           ((TOP_VALUE >> 2) + 1)
#define HALF                ((TOP_VALUE >> 1) + 1)
#define THIRD_QTR           (((3*TOP_VALUE) >> 2) + 1)

#define MAX_FREQ_MODEL      16384
#define MAX_FREQUENCY2      128 /* 16384 */
#define MAX_FREQUENCY1      256

#define NO_OF_SYMBOLS_0     3
#define NO_OF_GROUPS_0      16

#define NO_OF_SYMBOLS_1     (256 - 3)
#define NO_LEVEL_1          (7 + 1)

/*
 * ARI function
 */
#define OUTPUT_1()							\
{   buffer >>= 1;							\
    buffer |= 0x80;							\
	if (--bits_to_go == 0)					\
    {   *buf++ = buffer;					\
        bits_to_go = 8; }					\
}

#define OUTPUT_0()							\
{   buffer >>= 1;							\
	if (--bits_to_go == 0)					\
    {   *buf++ = buffer;					\
		bits_to_go = 8; }					\
}

#define BIT_PLUS_FOLLOW_1()					\
{											\
	OUTPUT_1()								\
	for (;bits_to_follow>0;--bits_to_follow)\
	  OUTPUT_0()							\
}

#define BIT_PLUS_FOLLOW_0()					\
{											\
	OUTPUT_0()								\
	for (;bits_to_follow>0;--bits_to_follow)\
	  OUTPUT_1()							\
}

#define INPUT_BIT(V)						\
{   if (bits_to_go == 0)					\
	{ buffer = *buf;						\
	buffer &= (buf >= length_buf) - 1;		\
	buf++;									\
    bits_to_go = 8; }						\
    bits_to_go--;							\
	V(buffer & 1);							\
	buffer >>= 1;							\
}

/*
 * ARI function
 */
#define ENCODE_SYMBOL_SM0(symbol, cumfreq, cumfreq0)					\
{	{uint32 range = high - low + 1;										\
	high = low + ((cumfreq[symbol - 1] * range) / cumfreq0) - 1;		\
	low += (cumfreq[symbol] * range) / cumfreq0; }						\
	for (;;)															\
    {	if (high < HALF)												\
			BIT_PLUS_FOLLOW_0()											\
		else if (low >= HALF)											\
{ BIT_PLUS_FOLLOW_1()  low -= HALF; high -= HALF; }						\
			 else if ((low >= FIRST_QTR) & (high < THIRD_QTR))			\
{ bits_to_follow++; low -= FIRST_QTR; high -= FIRST_QTR; }				\
				else break;												\
				low <<= 1;												\
				high <<= 1;												\
				high++; } }

#define ENCODE_SYMBOL_SM1(symbol, t, cumfreq0)							\
{	{uint32 range = high - low + 1;										\
	high = low + ((t[symbol - 1].cum * range) / cumfreq0) - 1;			\
	low += (t[symbol].cum * range) / cumfreq0; }						\
	for (;;)															\
    {	if (high < HALF)												\
			BIT_PLUS_FOLLOW_0()											\
		else if (low >= HALF)											\
{ BIT_PLUS_FOLLOW_1()  low -= HALF; high -= HALF; }						\
			 else if ((low >= FIRST_QTR) & (high < THIRD_QTR))			\
{ bits_to_follow++; low -= FIRST_QTR; high -= FIRST_QTR; }				\
				else break;												\
				low <<= 1;												\
				high <<= 1;												\
				high++; } }

/*
 * ARI function
 */
#define DECODE_SYMBOL_SM0(retour, cumfreq)								\
{	{ uint32 range = high - low + 1;									\
	uint32 cum = ((value - low + 1) * cumfreq[0] - 1) / range;			\
	retour = 0; while (cumfreq[++retour] > cum);						\
	high = low + ((cumfreq[retour - 1] * range) / cumfreq[0]) - 1;		\
	low += (cumfreq[retour] * range) / cumfreq[0]; }					\
	for (;;)															\
    {	if (high >= HALF)												\
		{ if (low >= HALF)												\
		  { value -= HALF; low -= HALF; high -= HALF; }					\
		      else if ((low >= FIRST_QTR) & (high < THIRD_QTR))			\
			  {value -= FIRST_QTR; low -= FIRST_QTR; high -= FIRST_QTR;}\
			    else break; }											\
		low <<= 1; high <<= 1; high++; value <<= 1;						\
		INPUT_BIT(value +=); } }

#define DECODE_SYMBOL_SM1(retour, t)									\
{	{ uint32 range = high - low + 1;									\
	uint32 cum = ((value - low + 1) * t[0].cum - 1) / range;			\
	retour = 0; while (t[++retour].cum > cum);							\
	high = low + ((t[retour - 1].cum * range) / t[0].cum) - 1;			\
	low += (t[retour].cum * range) / t[0].cum; }						\
	for (;;)															\
    {	if (high >= HALF)												\
		{ if (low >= HALF)												\
		  { value -= HALF; low -= HALF; high -= HALF; }					\
		      else if ((low >= FIRST_QTR) & (high < THIRD_QTR))			\
			  {value -= FIRST_QTR; low -= FIRST_QTR; high -= FIRST_QTR;}\
			    else break; }											\
		low <<= 1; high <<= 1; high++; value <<= 1;						\
		INPUT_BIT(value +=); } }


/*
 * ARI function
 */
#define START_OUTPUTING_BITS() { buffer = 0; bits_to_go = 8; }

#define DONE_OUTPUTING_BITS() {	*buf++ = buffer >> bits_to_go; }

#define START_ENCODING() { low = 0; high = TOP_VALUE; bits_to_follow = 0; }

#define DONE_ENCODING() { bits_to_follow++; if (low < FIRST_QTR) BIT_PLUS_FOLLOW_0() else BIT_PLUS_FOLLOW_1() }

#define START_INPUTING_BITS() { bits_to_go = 0; }

#define START_DECODING()						\
{   uint ii;	value = 0;						\
	for (ii = 1; ii <= CODE_VALUE_BITS; ++ii)	\
       INPUT_BIT(value = 2 * value +);			\
	low = 0; high = TOP_VALUE; }




static uint32 freq_tab_SM0[NO_OF_GROUPS_0][NO_OF_SYMBOLS_0] ALIGN;
static uint32 threshold[NO_OF_GROUPS_0] ALIGN;
static uint32 groups[NO_OF_SYMBOLS_0][NO_OF_SYMBOLS_0][NO_OF_SYMBOLS_0] ALIGN;
static uint32 tab_cum[NO_OF_SYMBOLS_0 + 1] ALIGN;

/*---------------------------------------------*/

/* Start the model (init) main function */
static
void Start_Model0(void)
{
  groups[0][0][0] = 0;
  threshold[0] = 128*MULT;

  groups[0][0][1] = 1;
  threshold[1] = 48*MULT;

  groups[0][1][0] = 2;
  groups[1][1][0] = 2;
  groups[1][0][0] = 2;
  threshold[2] = 64*MULT;

  groups[0][1][1] = 3;
  groups[1][1][1] = 3;
  threshold[3] = 64*MULT;

  groups[0][2][0] = 4;
  groups[2][0][0] = 4;
  groups[2][1][0] = 4;
  threshold[4] = 72*MULT;

  groups[1][2][1] = 5;
  groups[1][2][0] = 5;
  threshold[5] = 72*MULT;

  groups[2][2][1] = 6;
  groups[2][2][0] = 6;
  threshold[6] = 72*MULT;

  groups[2][0][1] = 7;
  groups[0][0][2] = 7;
  threshold[7] = 72*MULT;

  groups[0][1][2] = 8;
  groups[1][1][2] = 8;
  groups[2][0][2] = 8;
  threshold[8] = 72*MULT;

  groups[2][1][2] = 9;
  groups[0][2][2] = 9;
  groups[1][2][2] = 9;
  threshold[9] = 72*MULT;

  groups[0][2][1] = 10;
  groups[2][1][1] = 10;
  threshold[10] = 72*MULT;

  groups[1][0][2] = 11;
  threshold[11] = 64*MULT;

  groups[1][0][1] = 12;
  threshold[12] = 64*MULT;

  groups[2][2][2] = 13;
  threshold[13] = 128*MULT;

  memset(freq_tab_SM0, 0, sizeof(uint)*NO_OF_GROUPS_0*NO_OF_SYMBOLS_0);
}

/*---------------------------------------------*/

INLINE static
uint Get_Cum_Encode(uint *cum_freq,
                    uint freq1,
                    uint freq2,
                    uint freq3)
{
  freq1++;
  freq2++;
  freq3++;
  cum_freq[0] = freq1 + freq2 + freq3;
  cum_freq[3] = 0;

  if (freq1 >= freq2)
    { /* freq1 >= freq2 */
      if (freq2 >= freq3)
        { /* freq1 >= freq2 >= freq3 */
          cum_freq[1] = freq2 + freq3;
          cum_freq[2] = freq3;
          return (3<<4)|(2<<2)|1;
        }
      else
        { /* freq3 > freq2 */
          cum_freq[2] = freq2;
          if (freq3 >= freq1)
            { /* freq3 >= freq1 >= freq2 */
              cum_freq[1] = freq1 + freq2;
              return (1<<4)|(3<<2)|2;
            }
          else
            { /* freq1 > freq3 >= freq2 */
              cum_freq[1] = freq3 + freq2;
              return (2<<4)|(3<<2)|1;
            }
        }
    }
  else
    { /* freq2 > freq1 */
      if (freq1 >= freq3)
        { /* freq2 > freq1 > freq3 */
          cum_freq[1] = freq1 + freq3;
          cum_freq[2] = freq3;
          return (3<<4)|(1<<2)|2;
        }
      else
        { /* freq3 >= freq1 */
          cum_freq[2] = freq1;
          if (freq2 >= freq3)
            { /* freq2 > freq3 > freq1 */
              cum_freq[1] = freq3 + freq1;
              return (2<<4)|(1<<2)|3;
            }
          else
            { /* freq3 >= freq2 > freq1 */
              cum_freq[1] = freq2 + freq1;
              return (1<<4)|(2<<2)|3;
            }
        }
    }
}

/*---------------------------------------------*/

INLINE static
uint Get_Cum_Decode(uint *cum_freq,
                    uint freq1,
                    uint freq2,
                    uint freq3)
{
  freq1++;
  freq2++;
  freq3++;
  cum_freq[0] = freq1 + freq2 + freq3;
  cum_freq[3] = 0;

  if (freq1 >= freq2)
    { /* freq1 >= freq2 */
      if (freq2 >= freq3)
        { /* freq1 >= freq2 >= freq3 */
          cum_freq[1] = freq2 + freq3;
          cum_freq[2] = freq3;
          return (2<<4)|(1<<2)|0;
        }
      else
        { /* freq3 > freq2 */
          cum_freq[2] = freq2;
          if (freq3 >= freq1)
            { /* freq3 >= freq1 >= freq2 */
              cum_freq[1] = freq1 + freq2;
              return (1<<4)|(0<<2)|2;
            }
          else
            { /* freq1 > freq3 >= freq2 */
              cum_freq[1] = freq3 + freq2;
              return (1<<4)|(2<<2)|0;
            }
        }
    }
  else
    { /* freq2 > freq1 */
      if (freq1 >= freq3)
        { /* freq2 > freq1 > freq3 */
          cum_freq[1] = freq1 + freq3;
          cum_freq[2] = freq3;
          return (2<<4)|(0<<2)|1;
        }
      else
        { /* freq3 >= freq1 */
          cum_freq[2] = freq1;
          if (freq2 >= freq3)
            { /* freq2 > freq3 > freq1 */
              cum_freq[1] = freq3 + freq1;
              return (0<<4)|(2<<2)|1;
            }
          else
            { /* freq3 >= freq2 > freq1 */
              cum_freq[1] = freq2 + freq1;
              return (0<<4)|(1<<2)|2;
            }
        }
    }
}

/*---------------------------------------------*/



/* Compress a block with an order-0 Structured Model */
uint32 Zip_SM0(uint32 len,
               uint8  *bufin)
{
  bool   in_run = true;
  uint8  *buf;
  uint32 ch1 = 0, ch2 = 0, ch3 = 0, i = 0;
  uint32 buffer, low, high, bits_to_go;
  sint32 bits_to_follow;

  buf = block.buffer;

  Start_Model0();
  START_OUTPUTING_BITS();
  START_ENCODING();

  while (i < len)
    {
      uint32 group, *freq_tab, symbol, ch;

      ch = bufin[i++];

      group = groups[ch3][ch2][ch1];
      freq_tab = freq_tab_SM0[group];

      symbol = Get_Cum_Encode(tab_cum, freq_tab[0], freq_tab[1], freq_tab[2]);
      symbol = (symbol >> (ch * 2)) & 3;
      ENCODE_SYMBOL_SM0(symbol, tab_cum, tab_cum[0]);

      ch3 = ch2;
      ch2 = ch1;
      ch1 = ch;

      if ((ch | in_run | group) == 0)
        {
          in_run = true;
          freq_tab[0] >>= 1;
          freq_tab[0] += MULT + 1;
          freq_tab[1] >>= 1;
          freq_tab[2] >>= 1;
        }
      else
        {
          in_run = (ch != 0) ? false : in_run;

          if (tab_cum[0] > ((in_run == 1) ? 8 * 1024 : threshold[group]))
            {
              freq_tab[0] >>= 1;
              freq_tab[1] >>= 1;
              freq_tab[2] >>= 1;
            }

          freq_tab[ch] += MULT;
        }
    }

  DONE_ENCODING();
  DONE_OUTPUTING_BITS();

  return (buf - block.buffer);
}



/*---------------------------------------------*/

/* Uncompress a block with an order-0 Structured Model */
void Unzip_SM0(uint32 len,
               uint8  *bufout)
{
  bool   in_run = true;
  uint8  *buf, *length_buf;
  uint32 ch1 = 0, ch2 = 0, ch3 = 0, i = 0;
  uint32 buffer = 0, low, high, value, bits_to_go;

  buf = block.buffer;
  length_buf = block.buffer_length;

  Start_Model0();
  START_INPUTING_BITS();
  START_DECODING();

  while (i < len)
    {
      uint32 group, *freq_tab, symbol, ch, order;

      group = groups[ch3][ch2][ch1];
      freq_tab = freq_tab_SM0[group];

      order = Get_Cum_Decode(tab_cum, freq_tab[0], freq_tab[1], freq_tab[2]);

      DECODE_SYMBOL_SM0(symbol, tab_cum);
      ch = (order >> ((symbol - 1) * 2)) & 3;

      ch3 = ch2;
      ch2 = ch1;
      ch1 = ch;

      bufout[i++] = ch;

      if ((ch | in_run | group) == 0)
        {
          in_run = true;
          freq_tab[0] >>= 1;
          freq_tab[0] += MULT + 1;
          freq_tab[1] >>= 1;
          freq_tab[2] >>= 1;
        }
      else
        {
          in_run = (ch != 0) ? false : in_run;

          if (tab_cum[0] > ((in_run == 1) ? 8 * 1024 : threshold[group]))
            {
              freq_tab[0] >>= 1;
              freq_tab[1] >>= 1;
              freq_tab[2] >>= 1;
            }

          freq_tab[ch] += MULT;
        }
    }
}




/*---------------------------------------------*/

static uint32 cc[256] ALIGN;

/*
 * unBWT for a given block (from bufinout to bufinout)
 * uses 4*len bytes + bufinout
 * len must be < (1<<24) (16Mb)
 */

void BWT_Decoding(uint32 len,
                  uint32 first,
                  uint8  *bufinout,
                  uint32 *offset)
{
  memset(cc, 0, sizeof(uint32) * 256);

  /* bufinout is spread over the high bytes (0xFF000000) of offset */
  {
    uint32 i = 0;

    for (; i < len; ++i)
      offset[i] = ((uint32)bufinout[ i ]) << 24;
  }

  /* we use only the lower bytes (0x00FFFFFF) of offset to store the counts */
  {
    uint32 *olen, *o = offset;

    /* we skip 'first' */
    olen = offset + first;
    for (; o < olen; ++o)
      *o |= cc[*o >> 24]++;

    o++;
    olen = offset + len;
    for (; o < olen; ++o)
      *o |= cc[*o >> 24]++;

    offset[first] |= cc[offset[first] >> 24]++;
  }


  {
    uint32 k = 255, j = len;

    do cc[k] = (j -= cc[k]);
    while (--k != 0);
    cc[0] = 0;
  }


  {
    uint8  *b = bufinout + len - 1;
    uint32 j = first;

    while (b >= bufinout)
      j = cc[*(b--) = (offset[j] >> 24)] + (offset[j] & 0x00FFFFFFUL);
  }
}

/*---------------------------------------------*/

/* reverse a block, improve compression with some binary files */
void Reverse_Block(uint8 *bufin,
                   uint8 *bufin_end)
{

  bufin_end--;
  for (; bufin < bufin_end; ++bufin, --bufin_end)
    {
      uint t = *bufin;
      *bufin = *bufin_end;
      *bufin_end = t;
    }
}

/*---------------------------------------------*/



/*
 * 32 bits asm call trick : We transform relative adresses (following the CALL
 * opcode) into absolute ones to improve compression. With most of files, it
 * seems to be useless to do the same with JMP opcode.
 */

void Win32_Coding(uint8 *bufin,
                  uint8 *bufin_end)
{
  uint8 *bufin_start = bufin;

  bufin_end -= 6;
  for (; bufin < bufin_end; ++bufin)
    {
      if (*bufin == WIN32_ASM_CALL)
        {

          bufin++;
          *(uint32*)bufin += (uint32)(bufin - bufin_start);
          bufin += 3;

        }
    }
}



/*---------------------------------------------*/

/* The reverse operation */
void Win32_Decoding(uint8 *bufin,
                    uint8 *bufin_end)
{
  uint8 *bufin_start = bufin;

  bufin_end -= 6;
  for (; bufin < bufin_end; ++bufin)
    {
      if (*bufin == WIN32_ASM_CALL)
        {

          bufin++;
          *(uint32*)bufin -= (uint32)(bufin - bufin_start);
          bufin += 3;

        }
    }
}

/*---------------------------------------------*/



/* trick to compute an absolute value without any test/jump */
INLINE static
uint32 MyAbs(sint32 a)
{
  ssint64 s;
  s.s64 = a;
  return (s.d.l^s.d.h)-s.d.h;
}


/*
 * We try to find out if it's a "multimedia" (image, audio) block and what is
 * its interlacing, because raw "multimedia" files (24 bits images, 8/16 bits
 * mono/stereo audio files) are usually interlaced we try different interlacing
 * and calculate the mean distance.
 */
uint MM_Test(uint8 *bufin,
             uint8 *bufin_end)
{
  union
    {
      uint16 *b16;
      uint8  *b8;
    } b;
  uint32 len32 = bufin_end - bufin;
  uint8  *b8_end = bufin_end - 1;
  uint32 s1 = 0, s2 = 0;
  uint32 t1 = 0, t2 = 0, t3 = 0;

  for (b.b8 = bufin + 4; b.b8 < b8_end; b.b8 += 2)
    {
      s1 += MyAbs(*(b.b16) - *(b.b16-1)) >> 8;
      s2 += MyAbs(*(b.b16) - *(b.b16-2)) >> 8;
      t1 += MyAbs(*(b.b8+1) - *(b.b8-0));
      t1 += MyAbs(*(b.b8+0) - *(b.b8-1));
      t2 += MyAbs(*(b.b8+1) - *(b.b8-1));
      t2 += MyAbs(*(b.b8+0) - *(b.b8-2));
      t3 += MyAbs(*(b.b8+1) - *(b.b8-2));
      t3 += MyAbs(*(b.b8+0) - *(b.b8-3));
    }

  /*	printf("\n|Wav16:%.2f,%.2f|", (double)s1/len32, (double)s2/len32);
  	printf("Wav8:%.2f,%.2f,%.2f|", (double)t1/len32, (double)t2/len32, (double)t3/len32);*/

  /* these thresholds seem to work ;) */
  if (t1 < 13*len32) return 1;
  else if (t2 < 13*len32) return 2;
  else if (t3 < 14*len32) return 3;
  else if (s1 < 13*len32) return 4;
  else if (s2 < 17*len32) return 5;
  else return 0;
}



/*---------------------------------------------*/



/*
 * We do a delta-encoding on the block (according to the interlacing), it
 * improves compression with "multimedia" files.
 */
void MM_Coding(uint8 *bufin,
               uint8 *bufin_end)
{
  switch (block.mm_type)
    {
    case 1: /* interlacing 1:1 (byte) */
    {
      uint32  *b32 = (uint32*)bufin + 1;
      uint32  *b32_end = (uint32*)bufin_end + 1;
      uuint32 u1, u2;
      u1.u32 = *(uint32*)bufin;
      for (; b32 < b32_end; ++b32)
        {
          u2.u32 = *b32;
          u1.b.ll = u2.b.ll - u1.b.hh;
          u1.b.lh = u2.b.lh - u2.b.ll;
          u1.b.hl = u2.b.hl - u2.b.lh;
          u1.b.hh = u2.b.hh - u2.b.hl;
          *b32 = u1.u32;
          u1.u32 = u2.u32;
        }
    }
    break;
    case 2: /* interlacing 1:2 (byte) */
    {
      uint32  *b32 = (uint32*)bufin + 1;
      uint32  *b32_end = (uint32*)bufin_end + 1;
      uuint32 u1, u2;
      u1.u32 = *(uint32*)bufin;
      for (; b32 < b32_end; ++b32)
        {
          u2.u32 = *b32;
          u1.b.ll = u2.b.ll - u1.b.hl;
          u1.b.lh = u2.b.lh - u1.b.hh;
          u1.b.hl = u2.b.hl - u2.b.ll;
          u1.b.hh = u2.b.hh - u2.b.lh;
          *b32 = u1.u32;
          u1.u32 = u2.u32;
        }
    }
    break;
    case 3: /* interlacing 1:3 (byte) */
    {
      uint32  *b32 = (uint32*)bufin + 1;
      uint32  *b32_end = (uint32*)bufin_end + 1;
      uuint32 u1, u2;
      u1.u32 = *(uint32*)bufin;
      for (; b32 < b32_end; ++b32)
        {
          u2.u32 = *b32;
          u1.b.ll = u2.b.ll - u1.b.lh;
          u1.b.lh = u2.b.lh - u1.b.hl;
          u1.b.hl = u2.b.hl - u1.b.hh;
          u1.b.hh = u2.b.hh - u2.b.ll;
          *b32 = u1.u32;
          u1.u32 = u2.u32;
        }
    }
    break;
    case 4: /* interlacing 1:1 (word) */
    {
      uint32  *b32 = (uint32*)bufin + 1;
      uint32  *b32_end = (uint32*)bufin_end + 1;
      uuint32 u1, u2;
      u1.u32 = *(uint32*)bufin;
      for (; b32 < b32_end; ++b32)
        {
          u2.u32 = *b32;
          u1.w.l = u2.w.l - u1.w.h;
          u1.w.h = u2.w.h - u2.w.l;
          *b32 = u1.u32;
          u1.u32 = u2.u32;
        }
    }
    break;
    case 5: /* interlacing 1:2 (word) */
    {
      uint32  *b32 = (uint32*)bufin + 1;
      uint32  *b32_end = (uint32*)bufin_end + 1;
      uuint32 u1, u2;
      u1.u32 = *(uint32*)bufin;
      for (; b32 < b32_end; ++b32)
        {
          u2.u32 = *b32;
          u1.w.l = u2.w.l - u1.w.l;
          u1.w.h = u2.w.h - u1.w.h;
          *b32 = u1.u32;
          u1.u32 = u2.u32;
        }
    }
    break;
    }
}



/*---------------------------------------------*/

/* The reverse operation */
void MM_Decoding(uint8 *bufin,
                 uint8 *bufin_end)
{
  switch (block.mm_type)
    {
    case 1:
    {
      uint32  *b32 = (uint32*)bufin + 1;
      uint32  *b32_end = (uint32*)bufin_end + 1;
      uuint32 u1, u2;
      u1.u32 = *(uint32*)bufin;
      for (; b32 < b32_end; ++b32)
        {
          u2.u32 = *b32;
          u2.b.ll += u1.b.hh;
          u2.b.lh += u2.b.ll;
          u2.b.hl += u2.b.lh;
          u2.b.hh += u2.b.hl;
          *b32 = u2.u32;
          u1.u32 = u2.u32;
        }
    }
    break;
    case 2:
    {
      uint32  *b32 = (uint32*)bufin + 1;
      uint32  *b32_end = (uint32*)bufin_end + 1;
      uuint32 u1, u2;
      u1.u32 = *(uint32*)bufin;
      for (; b32 < b32_end; ++b32)
        {
          u2.u32 = *b32;
          u2.b.ll += u1.b.hl;
          u2.b.lh += u1.b.hh;
          u2.b.hl += u2.b.ll;
          u2.b.hh += u2.b.lh;
          *b32 = u2.u32;
          u1.u32 = u2.u32;
        }
    }
    break;
    case 3:
    {
      uint32  *b32 = (uint32*)bufin + 1;
      uint32  *b32_end = (uint32*)bufin_end + 1;
      uuint32 u1, u2;
      u1.u32 = *(uint32*)bufin;
      for (; b32 < b32_end; ++b32)
        {
          u2.u32 = *b32;
          u2.b.ll += u1.b.lh;
          u2.b.lh += u1.b.hl;
          u2.b.hl += u1.b.hh;
          u2.b.hh += u2.b.ll;
          *b32 = u2.u32;
          u1.u32 = u2.u32;
        }
    }
    break;
    case 4:
    {
      uint32  *b32 = (uint32*)bufin + 1;
      uint32  *b32_end = (uint32*)bufin_end + 1;
      uuint32 u1, u2;
      u1.u32 = *(uint32*)bufin;
      for (; b32 < b32_end; ++b32)
        {
          u2.u32 = *b32;
          u2.w.l += u1.w.h;
          u2.w.h += u2.w.l;
          *b32 = u2.u32;
          u1.u32 = u2.u32;
        }
    }
    break;
    case 5:
    {
      uint32  *b32 = (uint32*)bufin + 1;
      uint32  *b32_end = (uint32*)bufin_end + 1;
      uuint32 u1, u2;
      u1.u32 = *(uint32*)bufin;
      for (; b32 < b32_end; ++b32)
        {
          u2.u32 = *b32;
          u2.w.l += u1.w.l;
          u2.w.h += u1.w.h;
          *b32 = u2.u32;
          u1.u32 = u2.u32;
        }
    }
    break;
    }
}

/*---------------------------------------------*/



static const uint move[256] ALIGN =
  { 0,1,2,3,4,5,6,7,8,9,31,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,
    96,32,10,33,34,44,59,58,63,39,40,41,42,43,45,35,46,47,48,49,50,51,52,53,54,55,56,
    57,37,36,60,62,61,38,64,65,70,71,72,66,74,73,75,67,83,84,77,79,80,68,81,82,76,78,
    85,69,87,86,88,89,90,91,92,93,94,95,30,97,102,103,104,98,106,105,107,99,115,116,
    109,111,112,100,113,114,108,110,117,101,119,118,120,121,122,
    130,131,132,133,134,135,136,137,138,139,140,141,142,143,144,145,146,147,148,149,
    150,151,152,153,154,155,156,157,158,159,160,161,162,163,164,165,166,167,168,169,170,
    171,172,173,174,175,176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,
    192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,209,210,211,212,
    213,214,215,216,217,218,219,220,221,222,223,224,225,226,227,228,229,230,231,232,233,
    234,235,236,237,238,239,240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,
    255,123,124,125,126,127,128,129
  };

/*
 * some ASCII transformations: alphabet reordering, carriage-return tagging,
 * upper-case letter tagging
 */
uint32 Filter1(uint8  *bufin,
               uint8  *bufout,
               uint32 len)
{
  uint  b0, b1;
  uint8 *bout = bufout, *blen = bufin + len;

  b1 = *bufin++;
  while (bufin < blen)
    {
      b0 = b1;
      b1 = *bufin++;
      /* upper-case letter tagging */
      if (((b0 - 65) < 26) | (b0 == TAG_CAPS))
        {
          *bout++ = 96;/* move[TAG_CAPS]; */
          b0 += 32;
        }

      *bout++ = move[b0];

      /* carriage-return tagging */
      if ((b0 == ASCII_CR) && ((b1 == ASCII_SP) | ((b1 - 97) < 26)))
        {
          *bout++ = 10;/* move[32]; */
        }
    }
  *bout++ = move[b1];

  return (bout - bufout);
}



/*---------------------------------------------*/

static const uint unmove[256] ALIGN =
  { 0,1,2,3,4,5,6,7,8,9,32,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,
    27,28,29,96,10,31,33,34,45,59,58,63,39,40,41,42,43,35,44,46,47,48,49,50,51,52,53,
    54,55,56,57,37,36,60,62,61,38,64,65,69,73,79,85,66,67,68,71,70,72,82,76,83,77,78,
    80,81,74,75,84,87,86,88,89,90,91,92,93,94,95,30,97,101,105,111,117,98,99,100,103,
    102,104,114,108,115,109,110,112,113,106,107,116,119,118,120,121,122,249,250,251,
    252,253,254,255,123,124,125,126,127,128,129,130,131,132,133,134,135,136,137,138,
    139,140,141,142,143,144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,
    159,160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,176,177,178,
    179,180,181,182,183,184,185,186,187,188,189,190,191,192,193,194,195,196,197,198,
    199,200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,215,216,217,218,
    219,220,221,222,223,224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,
    239,240,241,242,243,244,245,246,247,248
  };

/* The reverse operation */
uint32 UnFilter1(uint8  *bufin,
                 uint32 len)
{
  uint  b0, b1;
  uint8 *bin = bufin, *blen = bufin + len, *bout = bufin;

  /*	int i, j;
  	printf("\n");
  	for(i = 0; i < 256+1; ++i)
  	{
  		j = 0;
  		while(move[j] != i) j++;
  		printf("%u,", j);
  	}
  	printf("\n");*/

  b1 = unmove[*bin++];
  while (bin < blen)
    {
      b0 = b1;
      b1 = unmove[*bin++];

      if (b0 != TAG_CAPS)
        {
          *bout++ = b0;
          if ((b0 == ASCII_CR) && (b1 == ASCII_SP)) b1 = unmove[*bin++];
        }
      else
        {
          *bout++ = b1 - 32;
          b1 = unmove[*bin++];
        }
    }
  *bout++ = b1;

  return (bout - bufin);
}

/*---------------------------------------------*/



/*
 * Some phrase replacements.
 * It's far from being optimal, but I can't find any better substitution.
 */
uint32 Filter2(uint8 *bufin,
               uint8 *bufin_end)
{
  uint8  *buf, *bufout = bufin;

  for (buf = bufin; buf < (bufin_end - 3); ++buf)
    {
      uint32 b = *(uint32*)buf;
      switch (b & 0x00FFFFFF)
        {
        case 6450037 /* the */:
          *bufout++ = 129;
          buf+=2;
          break;
        case 6644857 /* you */:
          *bufout++ = 128;
          buf+=2;
          break;
        case 7103594 /* for */:
          *bufout++ = 127;
          buf+=2;
          break;
        case 6844513 /* and */:
          *bufout++ = 126;
          buf+=2;
          break;
        case 7823723 /* hav */:
          *bufout++ = 125;
          buf+=2;
          break;
        default:
          switch (b & 0x0000FFFF)
            {
            case 28259 /* he */:
              *bufout++ = 124;
              buf++;
              break;
            case 24950 /* nd */:
              *bufout++ = 123;
              buf++;
              break;
            default :
              *bufout++ = *buf;
              break;
            }
          break;
        }
    }

  for (; buf < bufin_end; ++buf)
    *bufout++ = *buf;

  return (bufout-bufin);
}



/*---------------------------------------------*/

/* The reverse operation */
uint32 UnFilter2(uint8 *bufin,
                 uint8 *bufout,
                 uint8 *bufin_end)
{
  uint8  *bufout_sav = bufout;

  for (; bufin < bufin_end; ++bufin)
    {
      switch (*bufin)
        {
        case 129 :
          *(uint32*)bufout = 6450037;
          bufout += 3;
          break;
        case 128 :
          *(uint32*)bufout = 6644857;
          bufout += 3;
          break;
        case 127 :
          *(uint32*)bufout = 7103594;
          bufout += 3;
          break;
        case 126 :
          *(uint32*)bufout = 6844513;
          bufout += 3;
          break;
        case 125 :
          *(uint32*)bufout = 7823723;
          bufout += 3;
          break;
        case 124 :
          *(uint16*)bufout = 28259;
          bufout += 2;
          break;
        case 123 :
          *(uint16*)bufout = 24950;
          bufout += 2;
          break;
        default  :
          *bufout++ = *bufin;
        }
    }

  return (bufout - bufout_sav);
}

/*---------------------------------------------*/



static uint32 tab[257] ALIGN;

/* Block analysis */
void Analysis(uint8 *bufin,
              uint8 *bufin_end)
{
  uint32 buf_len = (uint32)(bufin_end - bufin);
  uint32 nb_win32 = 0, above = 0, below = 0, nb_he = 0, max;
  uint32 count = 0, max_count = 0;
  uint8  *buf;
  sint32 i;

  memset(tab, 0, sizeof(uint32) * 257);

  for (buf = bufin; buf < bufin_end; ++buf)
    {
      uint b = *(uint16*)buf;

      nb_he += (b == ASCII_H_E);

      if ((b & 0xFF) == (b >> 8)) count++;
      else
        {
          if (count > max_count) max_count = count;
          count = 0;
        }

      b &= 0xFF;

      if (b == WIN32_ASM_CALL && *(buf + 4) == 0) nb_win32++;

      tab[b]++;
    }
  max_count = MAX(count, max_count);

  if (max_count > RUN_LENGTH_MAX) block.rle_encoding = true;
  else block.rle_encoding = false;

  /* look for a free char */
  i = 0;
  while (tab[i] != 0) i++;

  /* decide if it is a binary block (BIN type) */
  if ((i > 255)
      && (block.type == NO_TYPE)
      && (block.compression_mode != 0))
    {
      block.type = BIN;
    }

  /* find max char value */
  i = 255;
  while (tab[i] == 0) i--;
  max = i;

  /* count the chars out of the ASCII [10-127] set */
  for (i = 0; i < 10; ++i)
    below += tab[i];
  for (i = 128; i < 256; ++i)
    above += tab[i];

  /* decide if it is a text block (TEXT type) */
  if (((block.compression_mode & 1) == 1)
      && (block.type == NO_TYPE)
      && (max > 32)
      && ((above + below) < (buf_len >> 4))) /* no more than 6% out of the ASCII set */
    {
      block.type = TEXT;
    }

  if (((nb_he*1000) > (4*buf_len))
      && (buf_len > 128 * 1024)
      && (block.type == TEXT)
      && (max < 256-7))
    {
      block.english_encoding = true;
    }
  else
    {
      block.english_encoding = false;
    }

  if ((block.compression_mode != 0)
      && ((block.type == NO_TYPE) | (block.type == BIN))
      && ((tab[0xE8] * 128) > (uint32)(bufin_end - bufin))
      && ((nb_win32 * 6) > tab[0xE8]))
    /* if we have more than 1/6 of 0xE8,xx,xx,00 -> it's a win32 file (WIN_EXE type) */
    {
      block.type = WIN_EXE;
    }
}



/*---------------------------------------------*/



/*
 * split the block into two, one with the charset [0..2], and the other with
 * the charset [2..255]. By the way, it finds the max char value.
 */
uint32 Split(uint8 *bufin,
             uint8 *bufin_end,
             uint8 *bufout2)
{
  uint8 *sav_bufout2 = bufout2;

  block.mtf_max_char = 1;

  for (; bufin < bufin_end; ++bufin)
    if (*bufin >= 2)
      {
        block.mtf_max_char = MAX(*bufin, block.mtf_max_char);
        *bufout2++ = *bufin;
        *bufin = 2;
      }

  return (bufout2 - sav_bufout2);
}


/*---------------------------------------------*/

/* The reverse operation */
void UnSplit(uint8 *bufin,
             uint8 *bufin_end,
             uint8 *bufin2)
{
  for (; bufin < bufin_end; ++bufin)
    if (*bufin == 2) *bufin = *bufin2++;
}

/*---------------------------------------------*/

static const uint32 crc32_table[] ALIGN =
{
  0UL, 16777216UL, 33554432UL, 50331648UL, 67108864UL, 83886080UL, 100663296UL,
  117440512UL, 134217728UL, 150994944UL, 167772160UL, 184549376UL, 201326592UL,
  218103808UL, 234881024UL, 251658240UL, 268435456UL, 285212672UL, 301989888UL,
  318767104UL, 335544320UL, 352321536UL, 369098752UL, 385875968UL, 402653184UL,
  419430400UL, 436207616UL, 452984832UL, 469762048UL, 486539264UL, 503316480UL,
  520093696UL, 536870912UL, 553648128UL, 570425344UL, 587202560UL, 603979776UL,
  620756992UL, 637534208UL, 654311424UL, 671088640UL, 687865856UL, 704643072UL,
  721420288UL, 738197504UL, 754974720UL, 771751936UL, 788529152UL, 805306368UL,
  822083584UL, 838860800UL, 855638016UL, 872415232UL, 889192448UL, 905969664UL,
  922746880UL, 939524096UL, 956301312UL, 973078528UL, 989855744UL, 1006632960UL,
  1023410176UL, 1040187392UL, 1056964608UL, 1073741824UL, 1090519040UL, 1107296256UL,
  1124073472UL, 1140850688UL, 1157627904UL, 1174405120UL, 1191182336UL, 1207959552UL,
  1224736768UL, 1241513984UL, 1258291200UL, 1275068416UL, 1291845632UL, 1308622848UL,
  1325400064UL, 1342177280UL, 1358954496UL, 1375731712UL, 1392508928UL, 1409286144UL,
  1426063360UL, 1442840576UL, 1459617792UL, 1476395008UL, 1493172224UL, 1509949440UL,
  1526726656UL, 1543503872UL, 1560281088UL, 1577058304UL, 1593835520UL, 1610612736UL,
  1627389952UL, 1644167168UL, 1660944384UL, 1677721600UL, 1694498816UL, 1711276032UL,
  1728053248UL, 1744830464UL, 1761607680UL, 1778384896UL, 1795162112UL, 1811939328UL,
  1828716544UL, 1845493760UL, 1862270976UL, 1879048192UL, 1895825408UL, 1912602624UL,
  1929379840UL, 1946157056UL, 1962934272UL, 1979711488UL, 1996488704UL, 2013265920UL,
  2030043136UL, 2046820352UL, 2063597568UL, 2080374784UL, 2097152000UL, 2113929216UL,
  2130706432UL, 2147483648UL, 2164260864UL, 2181038080UL, 2197815296UL, 2214592512UL,
  2231369728UL, 2248146944UL, 2264924160UL, 2281701376UL, 2298478592UL, 2315255808UL,
  2332033024UL, 2348810240UL, 2365587456UL, 2382364672UL, 2399141888UL, 2415919104UL,
  2432696320UL, 2449473536UL, 2466250752UL, 2483027968UL, 2499805184UL, 2516582400UL,
  2533359616UL, 2550136832UL, 2566914048UL, 2583691264UL, 2600468480UL, 2617245696UL,
  2634022912UL, 2650800128UL, 2667577344UL, 2684354560UL, 2701131776UL, 2717908992UL,
  2734686208UL, 2751463424UL, 2768240640UL, 2785017856UL, 2801795072UL, 2818572288UL,
  2835349504UL, 2852126720UL, 2868903936UL, 2885681152UL, 2902458368UL, 2919235584UL,
  2936012800UL, 2952790016UL, 2969567232UL, 2986344448UL, 3003121664UL, 3019898880UL,
  3036676096UL, 3053453312UL, 3070230528UL, 3087007744UL, 3103784960UL, 3120562176UL,
  3137339392UL, 3154116608UL, 3170893824UL, 3187671040UL, 3204448256UL, 3221225472UL,
  3238002688UL, 3254779904UL, 3271557120UL, 3288334336UL, 3305111552UL, 3321888768UL,
  3338665984UL, 3355443200UL, 3372220416UL, 3388997632UL, 3405774848UL, 3422552064UL,
  3439329280UL, 3456106496UL, 3472883712UL, 3489660928UL, 3506438144UL, 3523215360UL,
  3539992576UL, 3556769792UL, 3573547008UL, 3590324224UL, 3607101440UL, 3623878656UL,
  3640655872UL, 3657433088UL, 3674210304UL, 3690987520UL, 3707764736UL, 3724541952UL,
  3741319168UL, 3758096384UL, 3774873600UL, 3791650816UL, 3808428032UL, 3825205248UL,
  3841982464UL, 3858759680UL, 3875536896UL, 3892314112UL, 3909091328UL, 3925868544UL,
  3942645760UL, 3959422976UL, 3976200192UL, 3992977408UL, 4009754624UL, 4026531840UL,
  4043309056UL, 4060086272UL, 4076863488UL, 4093640704UL, 4110417920UL, 4127195136UL,
  4143972352UL, 4160749568UL, 4177526784UL, 4194304000UL, 4211081216UL, 4227858432UL,
  4244635648UL, 4261412864UL, 4278190080UL
};

/* Compute the CRC32 for a block. */
uint32 Crc32(uint8  *buffer,
             uint8  *buffer_end,
             uint32 crc)
{
  while (buffer < buffer_end)
    crc = (crc >> 8) ^ crc32_table[(crc ^ *buffer++) & 0xFF];

  return crc;
}

/*---------------------------------------------*/



/* RLE (variant) compression, useful for removing runs */
uint32 RLE_Coding(uint8 *bufin,
                  uint8 *bufout,
                  uint8 *bufend)
{
  uint32 c, last = 0, count = 0;
  uint8  *bufsav = bufout;

  while (bufin < bufend)
    {
      *bufout++ = (c = *bufin++);
      if (c != last)
        {
          count = 0;
          last = c;
        }
      else if (++count == 7)
        {
          count = 0;
          while ((*bufin == c) & (bufin < bufend))
            {
              count++;
              bufin++;
            }

          while (count != 0)
            {
              *bufout++ = c + (count & 1);
              count >>= 1;
            }
          if ((bufin == bufend)
              | (*bufin == ((c + 1) & 0xFF))
              | (*bufin == ((c + 2) & 0xFF)))
            {
              *bufout++ = (c + 2) & 0xFF;
            }
        }
    }

  return (bufout - bufsav);
}



/*---------------------------------------------*/

/* The reverse operation */
uint32 RLE_Decoding(uint8 *bufin,
                    uint8 *bufout,
                    uint8 *bufend)
{
  uint32 c, last = 0, count = 0;
  uint8  *bufsav = bufout;

  while (bufin < bufend)
    {
      *bufout++ = (c = *bufin++);
      if (c != last)
        {
          count = 0;
          last = c;
        }
      else if (++count == 7)
        {
          uint i = 0;
          count = 0;

          while ((*bufin == ((c + 1)&0xFF)) | (*bufin == c))
            count |= (uint32)((*bufin++ - ((c + 1)&0xFF) + 1)&0xFF) << (i++);

          if (*bufin == ((c + 2)&0xFF)) bufin++;

          for (; count > 0; --count) *bufout++ = c;
          count = 0;
        }
    }

  return (bufout - bufsav);
}

/*---------------------------------------------*/

static uint order[256] ALIGN;



/* MTF (in fact a "move one from front") encoding function */
void M1FF2_Coding(uint8 *bufinout,
                  uint8 *bufend)
{
  uint i, c, l, flag = 0;

  for (i = 0; i < 256; ++i) order[i] = i;

  for (; bufinout < bufend; ++bufinout)
    {
      c = *bufinout;
      l = 0;
      while (order[l] != c) l++;
      *bufinout = l;

      if (l > 1)
        {
          for (; l > 1; --l) order[l] = order[l - 1];
          order[1] = c;
          l = 1;
        }
      else if ((l == 1) & flag)
        {
          order[1] = order[0];
          order[0] = c;
        }
      flag = l;
    }
}



/*---------------------------------------------*/

#define MAX_INSERTSORT 20

#define VECSWAP(lb,ub,n)	\
{	uint16 **lb_v = (lb);	\
	uint16 **ub_v = (ub);	\
	uint32 n_v = (n);		\
	for(; n_v > 0; --n_v)	\
	{	uint16 *t;			\
		    t = *lb_v;		\
		*lb_v = *ub_v;		\
		*ub_v = t;			\
		lb_v++; ub_v--; } }

#define KEY(a) *(*(a)+depth)

/*---------------------------------------------*/

INLINE static
void ShellSort_depth_2(uint16 **lowerb,
                       uint16 **upperb)
{
  uint16 **i;

  if (upperb - lowerb > 10)
    {
      for (i = lowerb + 4; i <= upperb; ++i)
        {
          uint16 **j, *t;
          t = *i;
          for (j = i - 4; j >= lowerb; j -= 4)
            {
              uint16 *ba = *j + 2;
              uint16 *bb = t + 2;
              while (*bb == *ba)
                {
                  ba++;
                  bb++;
                }
              if (*bb > *ba) break;
              *(j + 4) = *j;
            }
          *(j + 4) = t;
        }
    }

  for (i = lowerb + 1; i <= upperb; ++i)
    {
      uint16 **j, *t;
      t = *i;
      for (j = i - 1; j >= lowerb; --j)
        {
          uint16 *ba = *j + 2;
          uint16 *bb = t + 2;
          while (*bb == *ba)
            {
              ba++;
              bb++;
            }
          if (*bb > *ba) break;
          *(j + 1) = *j;
        }
      *(j + 1) = t;
    }
}

/*---------------------------------------------*/

INLINE static
void ShellSort(uint16 **lowerb,
               uint16 **upperb,
               uint32 depth)
{
  uint16 **i;

  if (upperb - lowerb > 10)
    {
      for (i = lowerb + 4; i <= upperb; ++i)
        {
          uint16 **j =  i - 4;
          uint16  *t = *i;
          for (; j >= lowerb; j -= 4)
            {
              uint16 *ba = *j + depth;
              uint16 *bb = t + depth;
              while (*bb == *ba)
                {
                  ba++;
                  bb++;
                }
              if (*bb > *ba) break;
              *(j + 4) = *j;
            }
          *(j + 4) = t;
        }
    }

  for (i = lowerb + 1; i <= upperb; ++i)
    {
      uint16 **j =  i - 1;
      uint16  *t = *i;
      for (; j >= lowerb; --j)
        {
          uint16 *ba = *j + depth;
          uint16 *bb = t + depth;
          while (*bb == *ba)
            {
              ba++;
              bb++;
            }
          if (*bb > *ba) break;
          *(j + 1) = *j;
        }
      *(j + 1) = t;
    }
}

/*---------------------------------------------*/

#define MAX_STACK   1024
#define PUSH(a,b,c) { stack[sp].lb=(a);stack[sp].ub=(b);stack[sp].d=(c);sp++; }
#define  POP(a,b,c) { --sp;  a=stack[sp].lb;  b=stack[sp].ub;  c=stack[sp].d; }

typedef struct
  {
    uint16 **lb;
    uint16 **ub;
    uint32 d;
  } stack_s;

static stack_s stack[MAX_STACK] ALIGN;

/* classic Sedgewick ternary sort, tweaked a bit for speed */
INLINE static
void TernarySort(uint16 **lowerb_t,
                 uint16 **upperb_t,
                 uint32 depth_t)
{
  sint32 sp = 1;
  uint32 v, depth;
  uint16 **lb, **ub, **i, **j, **lowerb, **upperb;

  lowerb = lowerb_t;
  upperb = upperb_t;
  depth  = depth_t;

  while (sp > 0)
    {
      {
        sint32 r = upperb - lowerb + 1;

        if (r <= MAX_INSERTSORT)
          {
            if (r >= 2) ShellSort(lowerb, upperb, depth);
            POP(lowerb, upperb, depth);
            continue;
          }

        if (r > 64)
          { /* median-of-3 median-of-3 */
            uint v1, v2, v3;
            r >>= 3;
            lb = lowerb;
            {
              uint v_1, v_2, v_3;
              v_1  = KEY(lb);
              v_2 = KEY(lb+=r);
              v_3 = KEY(lb+=r);
              v1 = MEDIAN(v_1, v_2, v_3);
            }
            {
              uint v_1, v_2, v_3;
              v_1  = KEY(lb+=r);
              v_2 = KEY(lb+=r);
              v_3 = KEY(lb+=r);
              v2 = MEDIAN(v_1, v_2, v_3);
            }
            {
              uint v_1, v_2, v_3;
              v_1  = KEY(lb+=r);
              v_2 = KEY(lb+=r);
              v_3 = KEY(upperb);
              v3 = MEDIAN(v_1, v_2, v_3);
            }
            v = MEDIAN(v1, v2, v3);
          }
        else
          { /* median-of-3 */
            uint v1, v2, v3;
            v1 = KEY(lowerb);
            v2 = KEY(lowerb + (r >> 1));
            v3 = KEY(upperb);
            v  = MEDIAN(v1, v2, v3);
          }
      }

      i = lb = lowerb;
      j = ub = upperb;

      for (;;)
        {
          sint32 r;

          while (i <= j && (r = KEY(i) - v) <= 0)
            {
              if (r == 0)
                {
                  uint16 *t = *i;
                  *i = *lb;
                  *(lb++) = t;
                }
              i++;
            }
          while (i <= j && (r = KEY(j) - v) >= 0)
            {
              if (r == 0)
                {
                  uint16 *t = *j;
                  *j = *ub;
                  *(ub--) = t;
                }
              j--;
            }

          if (i > j) break;

          {
            uint16 *t = *i;
            *(i++) = *j;
            *(j--) = t;
          }
        }

      if (ub < lb)
        {
          depth += 2;
          continue;
        }

      {
        uint32 r;
        r=MIN((uint32)(lb-lowerb), (uint32)(i-lb));
        VECSWAP(lowerb, i-1, r);
      }

      {
        uint32 r;
        r=MIN((uint32)(ub-j), (uint32)(upperb-ub));
        VECSWAP(i, upperb, r);
      }

      /* sort the smallest partition    */
      /* to minimize stack requirements */
      {
        sint32 r1, r2, r3;

        r1 = (i - lb) - 1;
        r2 = (ub - i);
        r3 = (upperb - lowerb) + (lb - ub) - 1;

        if (r1 < r2)
          {	/* r1 < r2 */
            if (r3 <= r1)
              {	/* r3 <= r1 < r2 */
                PUSH(upperb - r2, upperb, depth);
                PUSH(lowerb, lowerb + r1, depth);
                lowerb += r1 + 1;
                upperb -= r2 + 1;
                depth += 2;
              }
            else if (r3 <= r2)
              {	/* r1 < r3 <= r2 */
                PUSH(upperb - r2, upperb, depth);
                PUSH(lowerb + r1 + 1, upperb - r2 - 1, depth + 2);
                upperb = lowerb + r1;
              }
            else
              {	/* r1 < r2 < r3 */
                PUSH(lowerb + r1 + 1, upperb - r2 - 1, depth + 2);
                PUSH(upperb - r2, upperb, depth);
                upperb = lowerb + r1;
              }
          }
        else
          {   /* r1 > r2 */
            if (r3 >= r1)
              {	/* r3 >= r1 > r2 */
                PUSH(lowerb + r1 + 1, upperb - r2 - 1, depth + 2);
                PUSH(lowerb, lowerb + r1, depth);
                lowerb = upperb - r2;
              }
            else if (r3 >= r2)
              {	/* r1 > r3 >= r2 */
                PUSH(lowerb, lowerb + r1, depth);
                PUSH(lowerb + r1 + 1, upperb - r2 - 1, depth + 2);
                lowerb = upperb - r2;
              }
            else
              {	/* r1 > r2 > r3 */
                PUSH(lowerb, lowerb + r1, depth);
                PUSH(upperb - r2, upperb, depth);
                lowerb += r1 + 1;
                upperb -= r2 + 1;
                depth += 2;
              }
          }
      }
    }
}

/*---------------------------------------------*/

/* we use only one table instead of two */
static union
  {
    uint16 **start[65536];
    uint32    size[65536];
  } m ALIGN;

/* BWT memory need : 6*len + 256ko */
void BWT_Coding(uint32 len,
                uint32 *first,
                uint8  *buffer8)
{
  /* some aliases */
  union
    {
      uint16 **pointer;
      uint8  *bufinout;
    } p;
  union
    {
      uint8  *buffer;
      uint16 *buffer16;
    } b;


  p.bufinout = buffer8;
  b.buffer   = (uint8*)ROUND32(buffer8 + (len + RUN_LENGTH_MAX) * sizeof(uint16*));

  GET_TSC(p1);

  memset(m.size, 0, sizeof(uint32) * 65536);
  {
    uint32 i, val = buffer8[0];
    buffer8[len] = 255;

    for (i = 0; i < len; ++i)
      {
        val = ((val & 0xFF) << 8) | buffer8[i + 1]; /* strip buffer over words */
        m.size[val]++;								/* update bucket sizes */
        b.buffer16[i] = val;						/* store word */
      }
  }
  memset(b.buffer16 + len, 0xFF, sizeof(uint16) * (RUN_LENGTH_MAX + 32));

  /* compute start offset of each bucket */
  {
    sint32 j;
    uint32 l = m.size[0], k;
    m.start[0] = p.pointer;

    for (j = 1; j < 65536; ++j)
      {
        k = m.size[j];
        m.start[j] = m.start[j - 1] + l;
        l = k;
      }
  }

  GET_TSC(p2);
  STAT_ADD_TIME(time_bwt1, p2, p1);
  STAT_ADD_SIZE(kb_bwt, len);

  /* finalize bucket sorting */
  {
    uint16 *bu = b.buffer16, *bulen = b.buffer16 + len;
    for (; bu < bulen; ++bu)
      *(m.start[*bu]++) = bu;
  }

  GET_TSC(p1);
  STAT_ADD_TIME(time_bwt2, p1, p2);

  /* finalize BWT sort by sorting each bucket */
  {
    sint32 j;
    uint16 **deb = p.pointer;
    uint32 l = m.start[0] - p.pointer;
    for (j = 0; j < 65536; ++j)
      {
        if (l > 1)
          {
            if (l <= (MAX_INSERTSORT+2)) ShellSort_depth_2(deb, deb + l - 1);
            else TernarySort(deb, deb + l - 1, 2);
          }
        deb += l;
        l = m.start[j + 1] - m.start[j]; /* width of next bucket */
      }
  }

  GET_TSC(p2);
  STAT_ADD_TIME(time_bwt3, p2, p1);

  {
    uint32 i = 0;
    uint16 *pp = p.pointer[0];

    while (pp != b.buffer16)
      {
        p.bufinout[i] = *(pp - 1) >> 8;
        pp = p.pointer[++i];
      }

    p.bufinout[i] = *(b.buffer16 + len - 1) >> 8;
    *first = i;
    i++;

    for (; i < len; ++i)
      p.bufinout[i] = *(p.pointer[i] - 1) >> 8;
  }

  GET_TSC(p1);
  STAT_ADD_TIME(time_bwt4, p1, p2);
}



/*---------------------------------------------*/

/* MTF decoding function */
void M1FF2_Decoding(uint8 *bufinout,
                    uint8 *bufend)
{
  uint i, c, l, flag = 0;

  for (i = 0; i < 256; ++i) order[i] = i;

  for (; bufinout < bufend; ++bufinout)
    {
      l = *bufinout;
      c = order[l];
      *bufinout = c;

      if (l > 1)
        {
          for (; l > 1; --l) order[l] = order[l - 1];
          order[1] = c;
          l = 1;
        }
      else if ((l == 1) & flag)
        {
          order[1] = order[0];
          order[0] = c;
        }
      flag = l;
    }
}

typedef struct
  {
    uint32 char_to_index;
    uint32 index_to_char;
  } s_SM1_i;

typedef struct
  {
    uint32 freq;
    uint32 cum;
  } s_SM1_f;

static s_SM1_i     SM1_i[NO_LEVEL_1 + 1]  ALIGN;
static s_SM1_i tab_SM1_i[NO_LEVEL_1][256] ALIGN;
static s_SM1_f     SM1_f[NO_LEVEL_1 + 1]  ALIGN;
static s_SM1_f tab_SM1_f[NO_LEVEL_1][256] ALIGN;





/* Start the model (init) sub-function */
static
void Start_Model_2(s_SM1_i *t_i,
                   s_SM1_f *t_f)
{
  uint i;

  for (i = 0; i < t_f[0].freq; ++i)
    {
      t_i[i].char_to_index = i + 1;
      t_i[i + 1].index_to_char = i;
    }

  t_f[0].cum = t_f[0].freq;
  for (i = 1; i <= t_f[0].freq; ++i)
    {
      t_f[i].freq = 1;
      t_f[i].cum = t_f[0].freq - i;
    }
}

/*---------------------------------------------*/




/* Start the model (init) main function */
static
void Start_Model(void)
{
  uint i, cum = 0, max, l;

  for (i = 0; i < NO_LEVEL_1; ++i)
    {
      SM1_i[i].char_to_index = i + 1;
      SM1_i[i + 1].index_to_char = i;
    }

  SM1_f[0 + 1].freq = 1;
  i = 1;

  max = block.mtf_max_char - 1;
  while (max > 0)
    {
      l = 1 << i;
      if (max < l)
        {
          i--;
          tab_SM1_f[i][0].freq += max;
          tab_SM1_f[i + 1][0].freq = 0;
          Start_Model_2(tab_SM1_i[i], tab_SM1_f[i]);
          max = 0;
        }
      else
        {
          max -= l;
          SM1_f[i + 1].freq = 1;
          tab_SM1_f[i][0].freq = l;
          Start_Model_2(tab_SM1_i[i], tab_SM1_f[i]);
        }
      i++;
    }

  SM1_f[0].freq = i;
  for (; i > 0; --i)
    {
      SM1_f[i].cum = cum;
      cum += SM1_f[i].freq;
    }
  SM1_f[0].cum = cum;
}


/* Update of the model for the specified symbol (structured model level 2) */
INLINE static
void Update_Model_2(uint    symbol,
                    s_SM1_i *t_i,
                    s_SM1_f *t_f)
{
  uint k, fs, cum;
  uint max = MIN(t_f[0].freq * MAX_FREQUENCY2, MAX_FREQ_MODEL);

  if (t_f[0].cum >= max)
    {
      cum = 0;
      for (k = t_f[0].freq; k > 0; --k)
        {
          t_f[k].cum = cum;
          cum += (t_f[k].freq = (t_f[k].freq + 1) >> 1);
        }
      t_f[0].cum = cum;
    }

  k = symbol - 1;
  fs = t_f[symbol].freq;
  while ((k > 0) && (fs == t_f[k].freq)) k--;

  if (++k < symbol)
    {
      uint ch_k, ch_symbol;
      ch_k = t_i[k].index_to_char;
      ch_symbol = t_i[symbol].index_to_char;
      t_i[k].index_to_char         = ch_symbol;
      t_i[symbol].index_to_char    = ch_k;
      t_i[ch_k].char_to_index      = symbol;
      t_i[ch_symbol].char_to_index = k;
    }

  t_f[k].freq++;
  while (k > 0) t_f[--k].cum++;
}


/* Update of the model (twice) for the specified symbol (structured model level 1) */
INLINE static
void Update_Model_1(const uint smbl)
{
  uint k, fs, cum, symbol = smbl;

  if (SM1_f[0].cum >= MAX_FREQUENCY1)
    {
      cum = 0;
      for (k = SM1_f[0].freq; k > 0; --k)
        {
          SM1_f[k].cum = cum;
          cum += (SM1_f[k].freq = (SM1_f[k].freq + 1) >> 1);
        }
      SM1_f[0].cum = cum;
    }

  k = symbol - 1;
  fs = SM1_f[symbol].freq;
  while ((k > 0) && (fs == SM1_f[k].freq)) k--;

  if (++k < symbol)
    {
      uint ch_k, ch_symbol;
      ch_k = SM1_i[k].index_to_char;
      ch_symbol = SM1_i[symbol].index_to_char;
      SM1_i[k].index_to_char         = ch_symbol;
      SM1_i[symbol].index_to_char    = ch_k;
      SM1_i[ch_k].char_to_index      = symbol;
      SM1_i[ch_symbol].char_to_index = k;
    }

  symbol = k;
  k--;
  fs++;
  while ((k > 0) && (fs == SM1_f[k].freq)) k--;

  if (++k < symbol)
    {
      uint ch_k, ch_symbol;
      ch_k = SM1_i[k].index_to_char;
      ch_symbol = SM1_i[symbol].index_to_char;
      SM1_i[k].index_to_char         = ch_symbol;
      SM1_i[symbol].index_to_char    = ch_k;
      SM1_i[ch_k].char_to_index      = symbol;
      SM1_i[ch_symbol].char_to_index = k;
    }

  SM1_f[k].freq++;
  SM1_f[symbol].freq++;
  while (symbol > k) SM1_f[--symbol].cum++;
  while (symbol > 0) SM1_f[--symbol].cum += 2;
}

/*---------------------------------------------*/




/* Compress a block with an order-0 Structured Model */
uint32 Zip_SM1(uint32 len,
               uint8  *bufin)
{
  uint32 i = 0;
  uint32 buffer = 0, low, high, bits_to_go;
  sint32 bits_to_follow;
  uint8  *buf = NULL;

  buf = block.buffer;

  Start_Model();
  START_OUTPUTING_BITS();
  START_ENCODING();

  while (i < len)
    {
      uint j, ch, symbol;

      ch = bufin[i++] - 2;

      j = 0;
      if (ch != 0)
        {
          ch++;
          while ((ch >> ++j) != 1);
          if (tab_SM1_f[j][0].freq == 0) j--;
          ch -= 1 << j;
        }

      symbol = SM1_i[j].char_to_index;
      ENCODE_SYMBOL_SM1(symbol, SM1_f, SM1_f[0].cum);
      Update_Model_1(symbol);

      if (j != 0)
        {
          symbol = tab_SM1_i[j][ch].char_to_index;
          ENCODE_SYMBOL_SM1(symbol, tab_SM1_f[j], tab_SM1_f[j][0].cum);
          Update_Model_2(symbol, tab_SM1_i[j], tab_SM1_f[j]);
        }
    }

  DONE_ENCODING();
  DONE_OUTPUTING_BITS();

  return (buf - block.buffer);
}




#define    Crc32_2(a,b)		Crc32(a,b,0xFFFFFFFFUL)



static

sint32 CompressBlock(
  uint8  *buffer_in,
  uint32 len_in)

{
  bool   ff_bug = false;
  uint16 status;
  uint8  *buffer1, *buffer2, *buf_out1, *buf_out2 = NULL;
  uint32 len, len2, len_max, first, tot1 = 0, tot2, block_len;
  sint   err = 0;




  uint8 *sav_buffer_in = buffer_in;

  mem.buffer8 = (uint8*)MyMalloc(sizeof(uint8) * (len_in + RUN_LENGTH_MAX + 32) * 6);
  if (last_error != OK) return -1;
  buffer1 = mem.buffer8;
  buffer2 = (uint8*)ROUND32(buffer1 + len_in * 2);

  memcpy(buffer1, buffer_in, sizeof(uint8) * len_in);
  len = len_in;


  block_len = len;
  len_max = len;
  block.crc = Crc32_2(buffer1, buffer1 + len);
  block.mm_type = 0;

  STAT_ADD_SIZE(kb_tot, len);

  if (len > 64)
    {
      if (len < 3 * 1024) block.compression_mode = 0;

      /*- Beginning ---- Analyze ------------*/

      GET_TSC(p1);
      Analysis(buffer1, buffer1 + len);
      GET_TSC(p2);

      STAT_ADD_TIME(time_ana, p2, p1);
      STAT_ADD_SIZE(kb_ana, len);

      /*- End ---------- Analyze ------------*/

      /* trick for 'Canterbury Corpus: kennedy.xls' !, delta-encoding with record size of 13 */
      if (((uint32*)buffer1)[10] == 161480704 && ((uint32*)buffer1)[20] == 60818693)
        {
          uint  i;
          uint8 *b8_out = buffer2, *b8;

          for (i = 0; i < 13; ++i)
            for (b8 = buffer1 + 2320 + i; b8 < buffer1 + len; b8 += 13)
              *b8_out++ = *b8 - *(b8-13);
          memcpy(buffer1 + 2320, buffer2, len - 2320);
          block.rle_encoding = true;
          block.type = NO_TYPE;
          block.mm_type = 6;
        }

      /*- Beginning ---- RLE Coding ---------*/

      if (block.rle_encoding == true)
        {
          uint32 rle_len;

          GET_TSC(p1);
          rle_len = RLE_Coding(buffer1, buffer2, buffer1 + len);
          GET_TSC(p2);

          STAT_ADD_TIME(time_rle, p2, p1);
          STAT_ADD_SIZE(kb_rle, len);

          len = rle_len;
          memcpy(buffer1, buffer2, len);
        }

      /*- End ---------- RLE Coding ---------*/

      /*- Beginning ---- MM Coding ----------*/

      if ((len > 128 * 1024)
          & ((block.type == BIN) | (block.type == NO_TYPE))
          & (block.multimedia_test == true))
        {
          uint res;

          GET_TSC(p1);
          res = MM_Test(buffer1 + 512, buffer1 + len);
          GET_TSC(p2);

          STAT_ADD_TIME(time_ana, p2, p1);

          if (res != 0)
            {
              block.mm_type = res;
              block.type = MULTIMEDIA;
            }
        }

      if (block.mm_type != 0)
        {
          GET_TSC(p1);
          MM_Coding(buffer1, buffer1 + len);
          GET_TSC(p2);

          STAT_ADD_TIME(time_mm, p2, p1);
          STAT_ADD_SIZE(kb_mm, len);
        }

      /*- End ---------- MM Coding ----------*/

      /*- Beginning ---- W32/BIN Coding -----*/

      if (block.type == WIN_EXE)
        {
          GET_TSC(p1)
          Win32_Coding(buffer1, buffer1 + len);
          GET_TSC(p2);

          STAT_ADD_TIME(time_txt, p2, p1);
        }

      if (block.type == BIN || block.type == WIN_EXE)
        {
          GET_TSC(p1);
          Reverse_Block(buffer1, buffer1 + len);
          GET_TSC(p2);

          STAT_ADD_TIME(time_txt, p2, p1);
          STAT_ADD_SIZE(kb_txt, len);
        }

      /*- End ---------- W32/BIN Coding -----*/

      /*- Beginning ---- alpha/txt Cod. -----*/

      if (block.type == TEXT)
        {
          GET_TSC(p1);

          len = Filter1(buffer1, buffer2, len);

          memcpy(buffer1, buffer2, len);

          mem.buffer8 = (uint8*)MyRealloc(mem.buffer8, sizeof(uint8) * (len + RUN_LENGTH_MAX + 32) * 6);
          if (last_error != OK) return -1;

          buffer1 = mem.buffer8;

          len_max = MAX(len, len_max);

          GET_TSC(p2);

          STAT_ADD_TIME(time_txt, p2, p1);
          STAT_ADD_SIZE(kb_txt, len);
        }

      /*- End ---------- alpha/txt Cod. -----*/

      /*- Beginning ---- Phr. replacement ---*/

      if (block.english_encoding == true)
        {
          GET_TSC(p1);
          len = Filter2(buffer1, buffer1 + len);
          GET_TSC(p2);

          STAT_ADD_TIME(time_txt, p2, p1);
        }

      /*- End ---------- Phr. replacement ---*/

      /*- Beginning ---- BWT ----------------*/

      /* to avoid a bug if the block ends with a run of 0xFF */
      if (buffer1[len - 1] == 0xFF)
        {
          buffer1[len - 1] -= buffer1[len - 2];
          ff_bug = true;
        }

      BWT_Coding(len, &first, buffer1);

      if (last_error != OK) return -1;

      /*- End ---------- BWT ----------------*/

      /*- Beginning ---- MTF Coding ---------*/

      GET_TSC(p1);

      M1FF2_Coding(buffer1, buffer1 + len);

      /*- End ---------- MTF Coding ---------*/

      /*- Beginning ---- Split --------------*/

      buf_out1 = (uint8*)ROUND32(buffer1 + len);
      buffer2  = (uint8*)ROUND32(buf_out1 + len);
      buf_out2 = (uint8*)ROUND32(buffer2 + len);

      len2 = Split(buffer1, buffer1 + len, buffer2);

      GET_TSC(p2);

      STAT_ADD_TIME(time_mtf, p2, p1);
      STAT_ADD_SIZE(kb_mtf, len);

      /*- End ---------- Split --------------*/

      /*- Beginning ---- Arith Compression --*/

      block.buffer = buf_out1;

      GET_TSC(p1);
      tot1 = Zip_SM0(len, buffer1);
      GET_TSC(p2);

      STAT_ADD_TIME(time_st0, p2, p1);
      STAT_ADD_SIZE(kb_st0, len);

      if (last_error != OK) return -1;

      block.buffer = buf_out2;

      GET_TSC(p1);
      tot2 = Zip_SM1(len2, buffer2);
      GET_TSC(p2);

      STAT_ADD_TIME(time_st1, p2, p1);
      STAT_ADD_SIZE(kb_st1, len2);

      if (last_error != OK) return -1;

      /*- End ---------- Arith Compression --*/

      /* if we can't compress the block, we get the original block */
      if ((tot1+tot2) > len_max)
        {
          err = -1;



          mem.buffer8 = MyFree(mem.buffer8);
          mem.buffer8 = (uint8*)MyMalloc(sizeof(uint8) * len_in);
          if (last_error != OK) return -1;
          buf_out1 = mem.buffer8;

          memcpy(buf_out1, buffer_in, sizeof(uint8) * len_in);
          tot1 = len_in;


          len_max = tot1;
        }
    } /* (len > 50) */
  else
    {
      buf_out1 = buffer1;
      err = -1;
      tot1 = len;
    }

  /*- Beginning ---- Writings -----------*/

  /*  status
     *   1   1   1   1   1   1   1
   *   6   5   4   3   2   1   0   9   8   7   6   5   4   3   2   1
   * |err|                       |ffb|eng|   mm_type |      type |rle|
   */
  status = 0;
  status |= ff_bug;
  status <<= 1;
  status |= block.english_encoding;
  status <<= 3;
  status |= block.mm_type;
  status <<= 3;
  status |= block.type;
  status <<= 1;
  status |= block.rle_encoding;

  if (err == -1) status |= 32768;

  /* we add a small margin, in case... */
  len_max += 32;



  WRITE_M(&len_max, sizeof(len_max));
  WRITE_M(&block.crc, sizeof(block.crc));
  if ((status & 32768) == 0)
    {
      WRITE_M(&status, sizeof(status));
      WRITE_M(&len, sizeof(len));
      WRITE_M(&len2, sizeof(len2));
      WRITE_M(&first, sizeof(first));
      WRITE_M(&block.mtf_max_char, sizeof(block.mtf_max_char));
      WRITE_M(&tot2, sizeof(tot2));
      WRITE_M(buf_out2, sizeof(uint8) * tot2);
    }
  else
    {
      WRITE_M(&status, sizeof(status));
    }
  WRITE_M(&tot1, sizeof(tot1));
  WRITE_M(buf_out1, sizeof(uint8) * tot1);



  /*- End ---------- Writings -----------*/

  mem.buffer8 = MyFree(mem.buffer8);


  return (uint32)(buffer_in - sav_buffer_in);

}



int __stdcall __export ZZCompressBlock(unsigned char *buffer,
                                       unsigned int  size,
                                       unsigned int  compression_mode,
                                       unsigned int  multimedia_test)
{
  int len;

  last_error = OK;

  if (buffer == NULL
      || size > 20*1024*1024)
    {
      last_error = BAD_PARAMETER;
      return -1;
    }

  session = (session_param_s*)MyMalloc(sizeof(session_param_s));
  if (last_error != OK) return -1;

  block.type = NO_TYPE;
  session->multimedia_test = multimedia_test;
  session->compression_mode = compression_mode;

  len = CompressBlock(buffer, size);

  session = MyFree(session);

  return len;
}




static s_SM1_i     SM1_i[NO_LEVEL_1 + 1]  ALIGN;
static s_SM1_i tab_SM1_i[NO_LEVEL_1][256] ALIGN;
static s_SM1_f     SM1_f[NO_LEVEL_1 + 1]  ALIGN;
static s_SM1_f tab_SM1_f[NO_LEVEL_1][256] ALIGN;


/* Uncompress a block with an order-0 Structured Model */
void Unzip_SM1(uint32 len,
               uint8  *bufout)
{
  uint32 i = 0, buffer = 0, low, high, value, bits_to_go;
  uint8  *buf = NULL, *length_buf = NULL;

  buf = block.buffer;
  length_buf = block.buffer_length;

  Start_Model();
  START_INPUTING_BITS();
  START_DECODING();

  while (i < len)
    {
      uint ch, ch2, symbol;

      DECODE_SYMBOL_SM1(symbol, SM1_f);
      ch = SM1_i[symbol].index_to_char;
      Update_Model_1(symbol);

      if (ch > 0)
        {
          ch2 = ch;
          DECODE_SYMBOL_SM1(symbol, tab_SM1_f[ch2]);
          ch = tab_SM1_i[ch2][symbol].index_to_char;
          Update_Model_2(symbol, tab_SM1_i[ch2], tab_SM1_f[ch2]);
          ch += 1 << ch2;
          ch--;
        }

      bufout[i++] = ch + 2;
    }
}




static

sint32 UncompressBlock(modes mode,
                       uint8 *buffer_in)

{
  bool   ff_bug = false;
  uint16 status;
  uint32 len = 0, len2 = 0, len_max = 0, first = 0, tot1 = 0, tot2 = 0;
  uint8  *bufferin1, *bufferin2 = NULL, *bufferout = NULL;



  uint8 *sav_buffer_in = buffer_in;


  READ_M(&len_max, sizeof(len_max));
  READ_M(&block.crc, sizeof(block.crc));
  READ_M(&status, sizeof(status));

  if ((status & 32768) == 0)
    {
      ff_bug					= (status >> 8) & (1);
      block.english_encoding	= (status >> 7) & (1);
      block.mm_type			= (status >> 4) & (1 + 2 + 4);
      block.type				= (status >> 1) & (1 + 2 + 4);
      block.rle_encoding		= (status >> 0) & (1);
      READ_M(&len, sizeof(len));
      READ_M(&len2, sizeof(len2));
      READ_M(&first, sizeof(first));
      READ_M(&block.mtf_max_char, sizeof(block.mtf_max_char));

      mem.buffer8 = (uint8*)MyMalloc(sizeof(uint8) * len_max * 5 + 32 * 2);
      if (last_error != OK) return -1;
      bufferout = mem.buffer8;
      bufferin1 = (uint8*)ROUND32(bufferout + len_max);
      bufferin2 = (uint8*)ROUND32(bufferin1 + len_max);

      READ_M(&tot2, sizeof(tot2));
      READ_M(bufferin2, sizeof(uint8) * tot2);
    }
  else
    {
      mem.buffer8 = (uint8*)MyMalloc(sizeof(uint8) * len_max * 1);
      if (last_error != OK) return -1;
      bufferin1 = mem.buffer8;
    }
  READ_M(&tot1, sizeof(tot1));
  READ_M(bufferin1, sizeof(uint8) * tot1);

  buffer_in = sav_buffer_in;




  if ((status & 32768) == 0)
    {
      block.buffer        = bufferin1;
      block.buffer_length = bufferin1 + tot1;

      GET_TSC(p1);
      Unzip_SM0(len, bufferout);
      GET_TSC(p2);

      STAT_ADD_TIME(time_st0, p2, p1);
      STAT_ADD_SIZE(kb_st0, len);

      if (last_error != OK) return -1;

      block.buffer        = bufferin2;
      block.buffer_length = bufferin2 + tot2;

      GET_TSC(p1);
      Unzip_SM1(len2, bufferin1);
      GET_TSC(p2);

      STAT_ADD_TIME(time_st1, p2, p1);
      STAT_ADD_SIZE(kb_st1, len2);

      if (last_error != OK) return -1;

      GET_TSC(p1);

      UnSplit(bufferout, bufferout + len, bufferin1);

      M1FF2_Decoding(bufferout, bufferout + len);

      GET_TSC(p2);

      STAT_ADD_TIME(time_mtf, p2, p1);
      STAT_ADD_SIZE(kb_mtf, len);

      GET_TSC(p1);
      BWT_Decoding(len, first, bufferout, (uint32*)bufferin1);
      GET_TSC(p2);

      STAT_ADD_TIME(time_bwt1, p2, p1);
      STAT_ADD_SIZE(kb_bwt, len);

      if (last_error != OK) return -1;

      if (ff_bug == true)
        bufferout[len - 1] += bufferout[len - 2];

      if (block.english_encoding == true)
        {
          GET_TSC(p1);
          len = UnFilter2(bufferout, bufferin1, bufferout + len);
          memcpy(bufferout, bufferin1, len);
          GET_TSC(p2);

          STAT_ADD_TIME(time_txt, p2, p1);
        }

      if (block.type == TEXT)
        {
          GET_TSC(p1);
          len = UnFilter1(bufferout, len);
          GET_TSC(p2);

          STAT_ADD_TIME(time_txt, p2, p1);
          STAT_ADD_SIZE(kb_txt, len);
        }

      if (block.type == BIN || block.type == WIN_EXE)
        {
          GET_TSC(p1);
          Reverse_Block(bufferout, bufferout + len);
          GET_TSC(p2);

          STAT_ADD_TIME(time_txt, p2, p1);
          STAT_ADD_SIZE(kb_txt, len);
        }

      if (block.type == WIN_EXE)
        {
          GET_TSC(p1);
          Win32_Decoding(bufferout, bufferout + len);
          GET_TSC(p2);

          STAT_ADD_TIME(time_txt, p2, p1);
        }

      if (block.mm_type != 0)
        {
          GET_TSC(p1);
          MM_Decoding(bufferout, bufferout + len);
          GET_TSC(p2);

          STAT_ADD_TIME(time_mm, p2, p1);
          STAT_ADD_SIZE(kb_mm, len);
        }

      if (block.rle_encoding == true)
        {
          GET_TSC(p1);
          len = RLE_Decoding(bufferout, bufferin1, bufferout + len);
          memcpy(bufferout, bufferin1, len);
          GET_TSC(p2);

          STAT_ADD_TIME(time_rle, p2, p1);
          STAT_ADD_SIZE(kb_rle, len);
        }

      /* trick for 'Canterbury Corpus: kennedy.xls' !*/
      if (block.mm_type == 6)
        {
          uint  i;
          uint8 *b8_out, *b8, *bufferout2 = bufferin1;

          memcpy(bufferout2, bufferout, 2320);

          b8 = bufferout + 2320;
          for (i = 0; i < 13; ++i)
            for (b8_out = bufferout2 + 2320 + i; b8_out < bufferout2 + len; b8_out += 13)
              *b8_out = *b8++ + *(b8_out-13);

          memcpy(bufferout + 2320, bufferout2 + 2320, len - 2320);
        }

      STAT_ADD_SIZE(kb_tot, len);

      if (block.crc != Crc32_2(bufferout, bufferout + len))
        {
          last_error = CRC_ERROR;
          return -1;
        }

      if (mode == UNCOMPRESS)
        {

          WRITE_M(bufferout, sizeof(uint8) * len);

        }
    }
  else
    {
      STAT_ADD_SIZE(kb_tot, tot1);

      if (block.crc != Crc32_2(bufferin1, bufferin1 + tot1))
        {
          last_error = CRC_ERROR;
          return -1;
        }

      len = tot1;
      if (mode == UNCOMPRESS)
        {

          WRITE_M(bufferin1, sizeof(uint8) * len);

        }
    }

  mem.buffer8 = MyFree(mem.buffer8);



  return len;
}


int __stdcall __export ZZUncompressBlock(unsigned char *buffer)
{
  int len;

  last_error = OK;

  if (buffer == NULL)
    {
      last_error = BAD_PARAMETER;
      return -1;
    }

  session = (session_param_s*)MyMalloc(sizeof(session_param_s));
  if (last_error != OK) return -1;

  len = UncompressBlock(UNCOMPRESS, buffer);

  session = MyFree(session);

  return len;
}
